package data;

import exception.LocationException;
import java.io.File;
import java.util.*;
import java.util.function.Predicate;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;
import javax.xml.namespace.QName;
import xml.LocationValidationEventHandler;

/**
 * <p>Classe Java pour anonymous complex type.
 * 
 * <p>Le fragment de schéma suivant indique le contenu attendu figurant dans cette classe.
 * 
 * <pre>
 * &lt;complexType>
 *   &lt;complexContent>
 *     &lt;restriction base="{http://www.w3.org/2001/XMLSchema}anyType">
 *       &lt;sequence>
 *         &lt;element name="lieu" type="{http://www.w3.org/2001/XMLSchema}string"/>
 *         &lt;element name="tarif" type="{http://www.w3.org/2001/XMLSchema}integer"/>
 *         &lt;element name="reservations" type="{}periode" maxOccurs="unbounded"/>
 *         &lt;element name="genre">
 *           &lt;simpleType>
 *             &lt;restriction base="{http://www.w3.org/2001/XMLSchema}string">
 *               &lt;enumeration value="Studio"/>
 *               &lt;enumeration value="Appartement"/>
 *               &lt;enumeration value="Maison"/>
 *             &lt;/restriction>
 *           &lt;/simpleType>
 *         &lt;/element>
 *       &lt;/sequence>
 *     &lt;/restriction>
 *   &lt;/complexContent>
 * &lt;/complexType>
 * </pre>
 * 
 * 
 */
@XmlAccessorType(XmlAccessType.FIELD)
@XmlType(name = "", propOrder = {
    "lieu",
    "tarif",
    "reservations",
    "genre"
})
@XmlRootElement(name = "location")
public class Location {

    @XmlElement(required = true)
    protected String lieu;
    @XmlElement(required = true)
    protected int tarif;
    @XmlElement(required = true)
    protected List<Periode> reservations = new ArrayList();
    @XmlElement(required = true)
    protected Genre genre;
    public static enum Genre {
        Studio ("Studio"), 
        Appartement ("Appartement"), 
        Maison ("Maison");
        private final String genre;
        Genre(String genre) {
            this.genre=genre;
        }
    }
    private static List<Location> listeLocations = new ArrayList();

    public String getLieu() {
        return lieu;
    }

    public void setLieu(String lieu) {
        this.lieu = lieu;
    }

    public int getTarif() {
        return tarif;
    }

    public void setTarif(int tarif) {
        this.tarif = tarif;
    }

    public List<Periode> getReservations() {
        return reservations;
    }

    public void setReservations(List<Periode> reservations) {
        this.reservations = reservations;
    }

    public Genre getGenre() {
        return genre;
    }

    public void setGenre(Genre genre) {
        this.genre = genre;
    }
    
    public static void ajouterLocation(Location location) {
        Location.listeLocations.add(location);
    }
    
    public static void supprimerLocationByLieu(String lieu) {
        for (Iterator<Location> iloc = Location.listeLocations.iterator(); iloc.hasNext();) {
            if (iloc.next().getLieu().equals(lieu)) {
                Location.listeLocations.remove(iloc.next());
                return;
            }
        }
    }
    
    public static void modifierLocation(String lieu, int tarif, Genre genre) {
        for (Iterator<Location> iloc = Location.listeLocations.iterator(); iloc.hasNext();) {
            if (iloc.next().getLieu().equals(lieu)) {
                iloc.next().tarif = tarif;
                iloc.next().genre = genre;
                return;
            }
        }
    }
    
    public boolean isReservationPossible(Periode periode) {
        for (Iterator<Periode> iperiode = reservations.iterator(); iperiode.hasNext();) {
            if (iperiode.next().isChevauchement(periode)) return false;
        }
        return true;
    }
    
    public boolean isReservation(Periode periode) {
        for (Iterator<Periode> iperiode = reservations.iterator(); iperiode.hasNext();) {
            if (iperiode.next().equals(periode)) return true;
        }
        return false;
    }
    
    public boolean existsReservation() {
        return !reservations.isEmpty();
    }
    
    public static void ajouterReservation(Location loc, Periode periode) {
        if (loc.isReservationPossible(periode)) loc.reservations.add(periode);
    }
    
    public static void supprimerReservation(Location loc, Periode periode) {
        for (Iterator<Periode> iperiode = loc.reservations.iterator(); iperiode.hasNext();) {
            if (iperiode.next().equals(periode)) {
                loc.reservations.remove(periode);
                return;
            }
        }
    }
    
    public static void modifierReservation(Location loc, Periode periode1, Periode periode2) {
        for (Iterator<Periode> iperiode = loc.reservations.iterator(); iperiode.hasNext();) {
            if (iperiode.next().equals(periode1)) {
                iperiode.remove();
                loc.reservations.add(periode2);
                return;
            }
        }
    }
    
    public Location() {    
    }
    
    public Location(String lieu, int tarif, Genre genre) {
        this.lieu = lieu;
        this.tarif = tarif;
        this.genre = genre;
    }
    
    public static Location findLocationByLieu(String lieu) throws LocationException {
        Predicate<Location> isFound = loc -> loc.lieu.equals(lieu);
        try {
            Location loc = Location
                .listeLocations
                .stream()
                .filter(isFound)
                .findFirst()
                .get();
            return loc;
        } catch (NoSuchElementException e) {
            throw new LocationException("Location non définie pour le lieu " + lieu);
        }
    }
        
    public static OptionalDouble moyenneTarifLocation(String lieu) {
        Predicate<Location> isFound = loc -> loc.lieu.equals(lieu);
        OptionalDouble moyenne = Location
                .listeLocations
                .stream()
                .filter(isFound)
                .mapToDouble(loc -> loc.tarif)
                .average();
        return moyenne;
    }
        
        
    public static OptionalDouble moyenneTarifLocation(Genre genre) {
        Predicate<Location> isFound = loc -> loc.genre.equals(genre);
        OptionalDouble moyenne = Location
                .listeLocations
                .stream()
                .filter(isFound)
                .mapToDouble(loc -> loc.tarif)
                .average();
        return moyenne;
    }
        
    public static OptionalDouble moyenneTarifLocation() {
        OptionalDouble moyenne = Location
                .listeLocations
                .stream()
                .mapToDouble(loc -> loc.tarif)
                .average();
        return moyenne;
    }
        
    public static OptionalDouble nombreLieuxParTarif(int min, int max) {
        OptionalDouble nombre = Location
                .listeLocations
                .stream()
                .mapToDouble(loc -> loc.tarif)
                .filter(tarif -> tarif > min && tarif < max)
                .average();
        return nombre;
    }
        
    protected static JAXBContext jaxbContext = null;
    protected static Unmarshaller unmarshaller = null;

    public void XMLwrite(File file) throws LocationException {// question 16
        try {
            if (jaxbContext == null) 
                jaxbContext = JAXBContext.newInstance ("xml");
            if (unmarshaller == null) 
                unmarshaller = jaxbContext.createUnmarshaller();
    
            JAXBElement<Location> uneLocationJAXB = 
                  new JAXBElement<>(new QName("location"),Location.class,this);
            Marshaller marshaller = jaxbContext.createMarshaller();
            marshaller.marshal( uneLocationJAXB, file );
        } catch (JAXBException e) {
            throw new LocationException("Erreur d'écriture dans le fichier " 
                    + file.getName());
        }
    }

    public static void XMLread(File file) throws LocationException {// question 16
        try {
            if (jaxbContext == null) 
                jaxbContext = JAXBContext.newInstance ("xml");
            if (unmarshaller == null) 
                unmarshaller = jaxbContext.createUnmarshaller();

            unmarshaller.setEventHandler(new LocationValidationEventHandler());
            Location location = (Location) unmarshaller.unmarshal(file);
            try {
                Location.findLocationByLieu(location.getLieu());
                throw new LocationException("Lecture du fichier " + file.getName() + " : lieu déjà défini");
            }
            catch (LocationException lex) {
                Location.ajouterLocation(location);
            }
        } catch (JAXBException e) {
            throw new LocationException("Erreur de lecture du fichier " 
                    + file.getName());
        }
    }
    
    @Override
    public String toString() {
        return getLieu() + " (" + getGenre() + ") :" + getTarif() + "euros";
    }
    
}